from enum import Enum
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Coroutine,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    OrderedDict,
    Set,
    Tuple,
    Type,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
)


class CQMsgDict(TypedDict):
    """
    cq 消息 dict
    """

    type: str
    data: Dict[str, float | int | str]


class MsgNodeDict(TypedDict):
    """
    消息节点 dict
    """

    class CustomNodeData(TypedDict):
        """
        自定义消息节点 data dict
        """

        name: str
        uin: int
        content: List[CQMsgDict]

    class ReferNodeData(TypedDict):
        """
        引用消息节点 data dict
        """

        id: int

    type: Literal["node"]
    data: CustomNodeData | ReferNodeData


class ParseArgs:
    """
    命令参数类
    """

    def __init__(self, values: List[Any]) -> None:
        self.vals = values
        self.formatted = False


class User(int, Enum):
    """
    用户权限等级枚举
    """

    OWNER = 10000
    SU = 1000
    WHITE = 100
    USER = 10
    BLACK = -1


class PriorLevel(int, Enum):
    """
    优先级枚举。方便进行优先级比较，有 MIN, MAX, MEAN 三个枚举值
    """

    MIN = 0
    MAX = 1000
    MEAN = (MAX + MIN) // 2


class BotLife(Enum):
    """
    bot 生命周期枚举
    """

    LOADED = 1
    CONNECTED = 2
    BEFORE_CLOSE = 3
    BEFORE_STOP = 4
    EVENT_BUILT = 5
    ACTION_PRESEND = 6


class LogicMode(Enum):
    """
    逻辑模式枚举
    """

    AND = 1
    OR = 2
    NOT = 3
    XOR = 4

    @classmethod
    def calc(cls, logic: "LogicMode", v1: Any, v2: Any = None) -> bool:
        if logic == LogicMode.AND:
            return (v1 and v2) if v2 is not None else bool(v1)
        elif logic == LogicMode.OR:
            return (v1 or v2) if v2 is not None else bool(v1)
        elif logic == LogicMode.NOT:
            return not v1
        elif logic == LogicMode.XOR:
            return (v1 ^ v2) if v2 is not None else bool(v1)

    @classmethod
    def seq_calc(cls, logic: "LogicMode", values: List[Any]) -> bool:
        if len(values) <= 0:
            return False
        elif len(values) <= 1:
            return bool(values[0])

        idx = 0
        res = None
        while idx < len(values):
            if idx == 0:
                res = cls.calc(logic, values[idx], values[idx + 1])
                idx += 1
            else:
                res = cls.calc(logic, res, values[idx])
            idx += 1
        return res


T = TypeVar("T")


class Void:
    """
    表示无值，而不是 None 代表的“空值”
    """

    pass



import asyncio as aio

from ..types.abc import AbstractDispatcher
from ..types.exceptions import *
from ..types.typing import *

if TYPE_CHECKING:
    from ..context.session import BotSessionManager
    from ..plugin.bot import BotHookBus
    from ..plugin.handler import EventHandler
    from ..types.abc import BotEvent
    from ..utils.logger import Logger


class BotDispatcher(AbstractDispatcher):
    """
    bot 分发模块。负责将传递的事件分发到各事件通道
    （接收的事件类型：消息、请求、通知和元事件）
    """

    def __init__(
        self,
        channel_map: dict[str, tuple[Type["EventHandler"], ...]],
        bot_bus: Type["BotHookBus"],
        ctx_manager: Type["BotSessionManager"],
        logger: "Logger",
    ) -> None:
        super().__init__()
        self.handlers: Dict[Type["EventHandler"], List["EventHandler"]] = {}

        self.logger = logger
        self._channel_map = channel_map
        self._bot_bus = bot_bus
        self._ctx_managger = ctx_manager
        self._ready_signal = aio.Event()

        for tuple in self._channel_map.values():
            for channel in tuple:
                if channel not in self.handlers.keys():
                    self.handlers[channel] = []

    def add_handlers(self, all_handlers: List["EventHandler"]) -> None:
        """
        绑定事件处理器列表
        """
        for handler in all_handlers:
            self.handlers[handler.__class__].append(handler)
        for k in self.handlers.keys():
            self.handlers[k] = sorted(
                self.handlers[k], key=lambda x: x.priority, reverse=True
            )
        self._ready_signal.set()

    async def broadcast(self, event: "BotEvent", channel: Type["EventHandler"]) -> None:
        """
        向指定的通道推送事件
        """
        try:
            permit_priority = PriorLevel.MIN.value
            handlers = self.handlers[channel]
            for handler in handlers:
                # 事件处理器优先级不够，则不分配给它处理
                if handler.priority < permit_priority:
                    continue
                if handler._direct_rouse and (
                    await self._ctx_managger.try_attach(event, handler)
                ):
                    if handler.set_block and handler.priority > permit_priority:
                        permit_priority = handler.priority
                    continue
                # evoke 返回的值用于判断，事件处理器内部经过各种检查后，是否选择处理这个事件。
                if not (await handler.evoke(event)):
                    # 如果决定不处理，则会跳过此次循环（也就是不进行“可能存在的优先级阻断操作”）
                    continue
                if handler.set_block and handler.priority > permit_priority:
                    permit_priority = handler.priority
        except Exception as e:
            self.logger.error(f"bot dispatcher 抛出异常")
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_locals(locals()))

    async def dispatch(self, event: "BotEvent") -> None:
        """
        把事件分发到对应的事件通道
        """
        await self._ready_signal.wait()
        await self._bot_bus.emit(BotLife.EVENT_BUILT, event, wait=True)
        for channel in self._channel_map[event.type]:
            aio.create_task(self.broadcast(event, channel))



"""
MeloBot 是插件化管理、基于异步会话机制的 qbot 开发框架。

项目源码和文档，请参考：https://github.com/aicorein/melobot
"""

from .context import *
from .init import MeloBot
from .meta import MetaInfo
from .models import *
from .plugin import *
from .types import *
from .utils import *

__version__ = MetaInfo().VER



from .abc import (
    BotAction,
    BotChecker,
    BotEvent,
    BotLife,
    BotMatcher,
    SessionRule,
    ShareObjArgs,
)
from .atools import cooldown, lock, semaphore, this_dir, timelimit, to_async, to_coro
from .typing import PriorLevel, User



help_docs = \
"""该命令共有子命令：
 ● docs: 显示帮助（后跟 run、status、options 中的一个获取帮助）
 ● run: 运行画图
 ● status: 显示状态
 ● options: 显示配置信息
 ● set: 修改运行配置"""


run_docs = \
"""~draw#run#... 功能：运行画图。

参数格式：
(提词) [反向提词:内置的反向提词] [宽:512] [高:512] [步数:20] [采样器:DPM++ 2M SDE Karras] [cfg 相关值:7] [种子:随机值] [差异种子:随机值] [差异强度:0] [超分启动:0] [超分重绘幅度:0.7] [超分倍数:2] [超分算法:Latent] [超分步数:内部默认值] [临时生效的配置:空]

注意：
1. 括号为必选，方框为存在默认值，冒号后代表默认值。
2. 如果需要使用的方框参数，前面存在使用默认值的方框参数，此时这些参数不能省略，必须使用 / 占位。否则则可以省略。
3. 传递的任何参数，使用 # 隔开

非白名单用户的限制：
1. 正向提词字符串长度 (0, 1000]
2. 反向提词字符串长度 [0, 500]
3. 宽高限制 [100, 960]
4. 总像素不得超过 518400
5. 步数限制 [10, 60]
6. 无权使用高清修复或临时配置功能

⚠ 特别提醒：为防止滥用，已开启 nsfw 过滤机制"""


status_docs = \
"""~draw#status 功能：获取工作状态，包含：启动参数和当前状态。

参数格式：无参数"""


options_docs = \
"""~draw#options 功能：显示配置信息，包含：可用模型、可用 vae、可用 loras、可用采样器、可用超分算法、可用脚本。

参数格式：无参数"""


set_docs = \
"""~draw#set#... 功能：修改配置。（至少需要 su 权限，查看权限可使用：~auth）

参数格式：
(名称1) (值1) (名称2) (值2)...
注意：传递的任何参数，使用 # 隔开

常用的配置项名称：
sd_model_checkpoint, sd_vae, CLIP_stop_at_last_layers"""



import re
from copy import deepcopy
from itertools import chain, zip_longest

from ..types.exceptions import *
from ..types.typing import *

if TYPE_CHECKING:
    from ..types.abc import BotAction


def text_msg(
    text: str,
) -> CQMsgDict:
    """
    普通文本消息
    """
    return {"type": "text", "data": {"text": text}}


def face_msg(
    icon_id: int,
) -> CQMsgDict:
    """
    QQ 表情
    """
    return {"type": "face", "data": {"id": f"{icon_id}"}}


def audio_msg(
    url: str,
    timeout: int = None,
    magic: bool = False,
) -> CQMsgDict:
    """
    语音消息
    """
    base = {
        "type": "record",
        "data": {
            "file": url,
        },
    }
    if magic:
        base["data"]["magic"] = 1
    if timeout:
        base["data"]["timeout"] = str(timeout)
    return base


def at_msg(
    qqId: int | Literal["all"],
    notInName: str = None,
) -> CQMsgDict:
    """
    at 消息。
    at 所有人时，`qqId` 传 "all"
    """
    base = {
        "type": "at",
        "data": {
            "qq": str(qqId),
        },
    }
    if notInName:
        base["data"]["name"] = notInName
    return base


def share_msg(
    url: str,
    title: str,
    content: str = None,
    image: str = None,
) -> CQMsgDict:
    """
    链接分享卡片消息。
    `content` 为描述语
    """
    base = {
        "type": "share",
        "data": {
            "url": url,
            "title": title,
        },
    }
    if content:
        base["data"]["content"] = content
    if image:
        base["data"]["image"] = image
    return base


def music_msg(
    platType: Literal["qq", "163", "xm"],
    songId: str,
) -> CQMsgDict:
    """
    音乐分享卡片消息（专有平台）
    """
    return {"type": "music", "data": {"type": platType, "id": songId}}


def custom_music_msg(
    url: str,
    audio: str,
    title: str,
    content: str = None,
    image: str = None,
) -> CQMsgDict:
    """
    自定义音乐分享卡片。
    `url` 为主页或网站起始页
    """
    base = {
        "type": "music",
        "data": {
            "type": "custom",
            "url": url,
            "audio": audio,
            "title": title,
        },
    }
    if content:
        base["data"]["content"] = content
    if image:
        base["data"]["image"] = image
    return base


def image_msg(
    url: str,
    picType: Literal["flash", "show"] = None,
    subType: Literal[0, 1] = None,
    useCache: Literal[0, 1] = 1,
) -> CQMsgDict:
    """
    图片消息。
    `url`: 图片 url。可以为本地路径，如：`file:///C:/users/15742/desktop/QQ图片20230108225606.jpg`；也可以为网络 url；还可以为 image id。
    `picType`: flash 为闪照，show 为秀图，不填为普通图片。
    `subType`: 只出现在群聊，0 为正常图片，1 为表情包
    """
    base = {
        "type": "image",
        "data": {
            "file": url,
        },
    }
    if picType:
        base["data"]["type"] = picType
    if subType:
        base["data"]["subType"] = str(subType)
    if useCache:
        base["data"]["cache"] = str(useCache)
    return base


def reply_msg(
    messageId: int,
) -> CQMsgDict:
    """
    回复消息
    """
    return {
        "type": "reply",
        "data": {
            "id": str(messageId),
        },
    }


def touch_msg(
    qqId: int,
) -> CQMsgDict:
    """
    戳一戳消息
    """
    return {
        "type": "poke",
        "data": {
            "qq": str(qqId),
        },
    }


def touch_msg(
    qqId: int,
) -> CQMsgDict:
    """
    openshamrock 的戳一戳消息
    """
    return {
        "type": "touch",
        "data": {
            "id": str(qqId),
        },
    }


def tts_msg(
    text: str,
) -> CQMsgDict:
    """
    腾讯自带 tts 语音消息
    """
    return {
        "type": "tts",
        "data": {
            "text": text,
        },
    }


def custom_msg_node(
    content: str | CQMsgDict | List[CQMsgDict],
    sendName: str,
    sendId: int,
    seq: str = None,
) -> MsgNodeDict:
    """
    自定义消息节点构造方法。转化字符串、消息、消息段为消息节点
    """
    if isinstance(content, str):
        msgs = text_msg(content)
        if not isinstance(msgs, list):
            msgs = [msgs]
    elif isinstance(content, dict):
        msgs = [content]
    elif isinstance(content, list):
        temp = []
        for _ in content:
            if isinstance(_, list):
                temp.extend(_)
            else:
                temp.append(_)
        msgs = temp
    ret = {
        "type": "node",
        "data": {"name": sendName, "uin": str(sendId), "content": msgs},
    }
    if seq:
        ret["data"]["seq"] = seq
    return ret


def refer_msg_node(msgId: int) -> MsgNodeDict:
    """
    引用消息节点构造方法
    """
    return {"type": "node", "data": {"id": str(msgId)}}


def cq_filter_text(s: str) -> str:
    """
    从 cq 消息字符串中，获取纯净的 cq text 类型消息
    """
    regex = re.compile(r"\[CQ:.*?\]")
    return regex.sub("", s)


def cq_escape(text: str) -> str:
    """
    cq 码特殊字符转义
    """
    return (
        text.replace("&", "&amp;")
        .replace("[", "&#91;")
        .replace("]", "&#93;")
        .replace(",", "&#44;")
    )


def cq_anti_escape(text: str) -> str:
    """
    cq 码特殊字符逆转义
    """
    return (
        text.replace("&#44;", ",")
        .replace("&#93;", "]")
        .replace("&#91;", "[")
        .replace("&amp;", "&")
    )


def to_cq_arr(s: str) -> list[CQMsgDict]:
    """
    从 cq 消息字符串转换为 cq 消息段
    """

    def replace_func(m) -> str:
        s, e = m.regs[0]
        cq_texts.append(m.string[s:e])
        return "\u0000"

    cq_regex = re.compile(r"\[CQ:.*?\]")
    cq_texts = []
    no_cq_str = cq_regex.sub(replace_func, s)
    pure_texts = map(
        lambda x: f"[CQ:text,text={x}]" if x != "" else x,
        no_cq_str.split("\u0000"),
    )
    content: str = "".join(
        chain.from_iterable(zip_longest(pure_texts, cq_texts, fillvalue=""))
    )

    cq_entity = content.split("]")[:-1]
    content: list = []
    for e in cq_entity:
        _ = e.split(",")
        cq_type = _[0][4:]
        data = {}
        for param_pair in _[1:]:
            name, val = param_pair.split("=")
            if cq_type != "text":
                val = cq_anti_escape(val)
            if val.isdigit() or (len(val) >= 2 and val[0] == "-" and val[1:].isdigit()):
                data[name] = int(val)
                continue
            try:
                data[name] = float(val)
            except Exception:
                data[name] = val
        content.append({"type": cq_type, "data": data})
    return content


def to_cq_str(content: list[CQMsgDict]) -> str:
    """
    从 cq 消息段转换为 cq 消息字符串
    """
    if isinstance(content, str):
        return content
    msgs = []
    for item in content:
        if item["type"] == "text":
            msgs.append(item["data"]["text"])
            continue
        s = f"[CQ:{item['type']}"
        for k, v in item["data"].items():
            s += f",{k}={cq_escape(str(v))}"
        s += "]"
        msgs.append(s)
    return "".join(msgs)


def to_cq_str_action(action: "BotAction") -> "BotAction":
    """
    转化 action 携带的 message 字段转为 cq 字符串格式，并返回新的 action。
    支持的 action 类型有：msg_action 和 forward_action
    """

    def _format_msg_action(action: "BotAction") -> None:
        action.params["message"] = to_cq_str(action.params["message"])

    def _format_forward_action(action: "BotAction") -> None:
        for item in action.params["messages"]:
            if "id" in item["data"].keys():
                continue
            item["data"]["content"] = to_cq_str(item["data"]["content"])

    _action = deepcopy(action)
    if _action.type == "send_msg":
        _format_msg_action(_action)
    elif _action.type in ("send_private_forward_msg", "send_group_forward_msg"):
        _format_forward_action(_action)
    else:
        raise BotActionError("传入的 action 因类型不匹配，不可被 cq 序列化")
    return _action


def get_cq(content: list[CQMsgDict], cq_type: str) -> list[CQMsgDict]:
    """
    从 content 获取指定类型的 cq 消息 dict
    """
    return [item for item in content if item["type"] == cq_type]


def get_cq_params(
    content: list[CQMsgDict], cq_type: str, param: str, type: Type[T] = None
) -> List[Any]:
    """
    从当前 content 中获取指定类型 cq 消息的指定 param，以列表形式返回。
    当没有任何对应类型的 cq 消息时，为空列表。如果有对应类型 cq 消息，
    但是 param 不存在，则在列表中产生值 None

    可以指定 type 来强制转换类型
    """
    res = []
    for item in content:
        if item["type"] == cq_type:
            val = item["data"].get(param)
            res.append(val)
    if type is not None:
        res = list(map(lambda x: type(x), res))
    return res



import time
import os
import json
import functools
import asyncio as aio
from core.Executor import EXEC, AuthRole
from common import *
from common.Action import custom_msg_node
from common.Store import BOT_STORE
from common.Exceptions import *
from common.Typing import *
from templates.sdapi.api import SdApi
from templates.sdapi.api_help import help_docs, run_docs, status_docs, options_docs, set_docs


format_funcs = [str, str, int, int, int, str, float, int, int, float, lambda x: bool(int(x)), float, int, str, int, lambda x: json.loads(x)]

arg_names = ['prompt', 'negative_prompt', 'width', 'height', 'steps', 'sampler_name', 'cfg_scale', 'seed', 'subseed', 'subseed_strength', 'enable_hr', 'denoising_strength', 'hr_scale', 'hr_upscaler', 'hr_second_pass_steps', 'script_name', 'script_args', 'override_settings', 'limit_access']

arg_defaults = [-1, None, 512, 512, 20, 'DPM++ 2M SDE Karras', 7, None, None, None, None, None, None, None, None, None]

superior_ids = set(BOT_STORE.config.super_user + [BOT_STORE.config.owner] + \
                   BOT_STORE.config.white_list)
su_ids = set(BOT_STORE.config.super_user + [BOT_STORE.config.owner])

cq_image_path = '/home/melodyecho/Program/cqhttp/data/images'


def format_with_defaults(formatters: List[Callable], idx: int, arg: Any) -> Callable:
    if arg == '':
        raise ValueError
    elif arg == '/':
        return arg_defaults[idx]
    else:
        return formatters[idx](arg)


def open_api() -> SdApi:
    try:
        api = SdApi()
        BOT_STORE.logger.info("sdapi 已刷新本地配置缓存")
        return api
    except Exception as e:
        BOT_STORE.logger.warning(f"sdapi 初始化失败：{e.__str__()}")
        return None


def close(api: SdApi) -> None:
    api.session.close()


@EXEC.template(
    aliases=['画图', 'sd'],
    userLevel=AuthRole.USER,
    comment=help_docs,
    prompt='功能复杂，请参考具体的帮助信息',
    preLoad=(open_api, close),
    timeout=600
)
async def draw(session: BotSession, subcmd: str, *args) -> None:
    api: SdApi = EXEC.get_resource('draw')
    if api is None:
        await session.send("sdapi 状态异常，暂无法提供服务")
        return
    lock = EXEC.get_cmd_lock('draw')

    if subcmd == 'run':
        if lock.locked():
            await session.send(f"正在画另一张图呢，进度是 {round(float(api.progress), 3)*100}%, 等一会儿吧~")
            return
        async with lock:
            await draw_run(session, api, *args)
    elif subcmd == 'docs':
        await draw_docs(session, *args)
    elif subcmd == 'status':
        await draw_status(session, api)
    elif subcmd == 'options':
        await draw_options(session, api)
    elif subcmd == 'set':
        await draw_set(session, api, *args)
    else:
        raise BotCmdExecFailed("无效子命令")


async def draw_docs(session: BotSession, sub_name: str) -> None:
    if sub_name in ['run', 'status', 'options', 'set']:
        await session.send(globals()[sub_name + '_docs'])
    else:
        await session.send("不存在该子命令的帮助信息")


async def draw_run(session: BotSession, api: SdApi, *args) -> None:
    if len(args)<1 or len(args)>16:
        raise BotCmdExecFailed("提供的画图参数太少或太多")
    
    formatters = format_funcs.copy()[: len(args)]
    params = list(args)

    try:
        if params[0] in ('', '/'):
            await session.send("prompt 必须提供，不存在默认值，也不允许空值")
            return
        for i in range(len(params)):
            params[i] = format_with_defaults(formatters, i, params[i])
    except Exception as e:
        if len(params[i]) > 20:
            params[i] = params[i][:20] + '...'
        await session.send(f"参数格式化失败！对于参数 {arg_names[i]} 提供的值是 {params[i]}")
        return

    # 最后一个转为命名参数
    override_settings = None
    if len(params) == 16:
        override_settings = params[-1]
        params = params[:-1]

    try:
        await session.send("开始生成...")
        if session.event.msg.sender.id in superior_ids:
            req_func = functools.partial(api.txt2img, *params, override_settings=override_settings, limit_access=False)
        else:
            req_func = functools.partial(api.txt2img, *params)

        ret = await aio.get_running_loop().run_in_executor(None, req_func)
        images, draw_params, info = ret
        img_name = str(time.time())
        images[0].save(cq_image_path + img_name + '.png')

        await session.send_forward([
            custom_msg_node(
                image_msg(img_name + '.png', useCache=0),
                sendName="律汐 bot",
                sendId=BOT_STORE.meta.bot_id
            ),
            custom_msg_node(
                info.split(r'\n')[2],
                sendName="律汐 bot",
                sendId=BOT_STORE.meta.bot_id
            ),
            custom_msg_node(
                '正向提词：\n' + params[0],
                sendName="律汐 bot",
                sendId=BOT_STORE.meta.bot_id
            ),
            custom_msg_node(
                '反向提词：\n' + info.split(r'\n')[1][17:],
                sendName="律汐 bot",
                sendId=BOT_STORE.meta.bot_id
            )
        ])
        EXEC.callback(20, os.remove, cq_image_path + img_name + '.png')
    
    except SdApi.ApiBaseException as e:
        await session.send("异常抛出：\n" + e.err)
    except Exception as e:
        await session.send("异常抛出：\n" + e.__str__())


async def draw_status(session: BotSession, api: SdApi) -> None:
    try:
        endpoint = api.endpoint + '（ssh forward）'
        xformer_enable = api.store.cmd_args['xformers']
        opt_split_attention = api.store.cmd_args['opt_split_attention']
        disable_nan_check = api.store.cmd_args['disable_nan_check']
        nowebui = api.store.cmd_args['nowebui']
        no_half_vae = api.store.cmd_args['no_half_vae']
        no_half = api.store.cmd_args['no_half']
        model = api.store.options['sd_model_checkpoint']
        vae = api.store.options['sd_vae']
        clip = int(api.store.options['CLIP_stop_at_last_layers'])
        eta = api.store.options['eta_ancestral']
        eta_noise_seed_shift = int(api.store.options['eta_noise_seed_delta'])
        job_count, progress = api.job_count, round(float(api.progress), 3)*100

        status_str = f"远端 sd Server 状态： \n\
 ● 接口：{endpoint} \n\
 ● 正在进行的任务: {job_count}  {progress}%  \n\
 ● 当前模型：{model} \n\
 ● 当前 vae：{vae} \n\
 ● 当前 clip 步数：{clip} \n\
 ● eta：{eta} \n\
 ● eta 噪声种子偏移：{eta_noise_seed_shift} \n\
 ● 启用 xformers 优化：{xformer_enable}  \n\
 ● 启用跨注意层优化：{opt_split_attention}  \n\
 ● 禁用 nan 检查：{disable_nan_check}  \n\
 ● webui 同时启动：{not nowebui}  \n\
 ● 禁用半精度：{no_half}  \n\
 ● vae 上禁用半精度：{no_half_vae} \
"
        await session.send(status_str)
    except Exception as e:
        await session.send("异常抛出：\n" + e.__str__())


async def draw_options(session: BotSession, api: SdApi) -> None:
    lora_str_list = [lora_info[0] + f'（{lora_info[1]}）' for lora_info in api.store.loras]
    basic_sep = '\n'

    str_list = [
        f" ● 可用的模型：\n{ basic_sep.join(api.store.models) }",
        f" ● 可用的 vae：\n{ basic_sep.join(api.store.vaes) }",
        f" ● 可用的 loras（括号内为别称）：\n{ basic_sep.join(lora_str_list) }",
        f" ● 可用的采样器：\n{ basic_sep.join(api.store.samplers) }",
        f" ● 可用的超分算法：\n{ basic_sep.join(api.store.upsclaers) }",
        f" ● 可用的脚本：\n无（默认强制使用 nsfw 过滤脚本，其他不可用）"
    ]

    await session.send_forward([
        custom_msg_node(
            string,
            sendName="律汐 bot",
            sendId=BOT_STORE.meta.bot_id
        )
        for string in str_list
    ])


async def draw_set(session: BotSession, api: SdApi, *args) -> None:
    if session.event.msg.sender.id not in su_ids:
        return
    if len(args)%2 != 0:
        await session.send("配置名、值的个数不匹配")
        return

    names, vals = [], []
    for idx, arg in enumerate(args): 
        names.append(arg) if (idx+1)% 2 != 0 else vals.append(arg)

    try:
        for idx in range(len(vals)):
            vals[idx] = eval(vals[idx])
    except NameError:
        raise BotCmdExecFailed("任何配置项的值，请以 python 表达式格式给出，否则内部无法正确格式化类型")

    try:
        req_func = functools.partial(api.modify_option, dict(zip(names, vals)))
        refresh_func = functools.partial(api.refresh_store, api.store)
        await aio.get_running_loop().run_in_executor(None, req_func)
        await aio.get_running_loop().run_in_executor(None, refresh_func)
        await session.send("远端 sd 服务器配置已生效，本地缓存已更新")
    except SdApi.ApiBaseException as e:
        await session.send("异常抛出：\n" + e.err)
    except Exception as e:
        await session.send("异常抛出：\n" + e.__str__())


from api import SdApi


api = SdApi()
prompt = '(((masterpiece))),(((best quality))),((an extremely delicate and beautiful)),(((extremely detailed 8k wallpaper))),a sunny day,((whole body shot)),(in a grassland),(detailed grass),(((only one girl))),((a girl sitting on grass)),((look at viewer sideway)),(long hair over shoulder),((aqua gradient hair)),((crossed bangs)),hair between eyes,(white hair flower with green leaves),detailed face,(beautiful detailed eyes),(aqua eyes),detailed mouth and nose,(cute smile),light blush,small breasts,((aqua serafuku with a bow)),aqua skirt,((yokozuwari)),((slender legs)),white pantyhose,((brown loafers shoe)),beautiful and warm lighting'
negative_prompt = 'nsfw,lowres,bad anatomy,bad hands,((bad legs)),((three or four legs)),(bad feet),text,error,missing fingers,extra digit,fewer digits,cropped,worst quality,low quality,normal quality,jpeg artifacts,signature,watermark,username,blurry,bad mouth,(((two girls)))'
steps = 28
seed = 3736812062
subseed = 1016585624
strength = 0.1
cfg = 8
denosing = 0.4


images, params, info = api.txt2img(prompt, negative_prompt, steps=steps, cfg_scale=cfg, seed=seed, 
                                   subseed=subseed, subseed_strength=strength, limit_access=False)
images[0].save('./test.png')
print(info.split(r'\n')[2])


import io
import sys

from better_exceptions import ExceptionFormatter
from rich import print


class BotException(Exception):
    """
    bot 异常基类
    """

    def __init__(self, msg: str):
        super().__init__(self, msg)
        self.err = msg

    def __str__(self):
        return self.err


class PluginLoadError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginBuildError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotConnectFailed(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotHookError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class EventHandlerError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotCheckerError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotMatcherError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class TryFlagFailed(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ShareObjError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginSignalError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotActionError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotSessionError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotBaseUtilsError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgParseError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgFormatInitError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgVerifyFailed(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


class ArgLackError(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


class DirectRetSignal(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


class SessionHupTimeout(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


_EXC_FORMATTER = ExceptionFormatter(colored=False, pipe_char="|", cap_char="└")


def get_better_exc(e: Exception) -> str:
    """
    返回生成更好的异常字符串
    """
    return "".join(
        _EXC_FORMATTER.format_exception(e.__class__, e, sys.exc_info()[2])
    ).strip("\n")


def get_rich_locals(locals: dict) -> str:
    """
    返回使用 rich 格式化的 locals() dict
    """
    sio = io.StringIO()
    print(locals, file=sio)
    return sio.getvalue().strip("\n")



import structlog


processors = [
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.TimeStamper(
        fmt='%Y-%m-%d %H:%M:%S' if log_console else 'iso'),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
]
if log_console:
    processors.append(structlog.dev.ConsoleRenderer())
else:
    processors.append(structlog.processors.JSONRenderer())
structlog.configure(
    processors=processors,
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

# Lib logger to info level.
logger = structlog.get_logger('mylibrary')
logger.setLevel(logging._nameToLevel.get(
    self.config['log_level'], logging.INFO))

# Set handler and level to root logger.
logging.root.setLevel(logging.WARNING)
if log_console:
    logging.root.addHandler(logging.StreamHandler(sys.stdout))
else:
    logging.root.addHandler(
        logging.handlers.TimedRotatingFileHandler(
            self.config['logfile'], encoding='utf-8',
            # Rotate logfiles weekly on Monday, keep last 10 weeks.
            when='W0', atTime=datetime.time(0, 0, 0), backupCount=10))


from .checker import GroupMsgLvl, MsgCheckerGen, MsgLvlChecker, PrivateMsgLvl
from .formatter import ArgFormatter
from .matcher import ContainMatch, EndMatch, FullMatch, RegexMatch, StartMatch
from .parser import CmdParser, CmdParserGen



"""
以模块运行模式启动 bot，可以将 bot 程序包装启动。
从而可以实现重启等功能
"""

import os
import pathlib
import subprocess
import sys

sys.path.append("..")
from melobot.meta import (
    EXIT_CLOSE,
    EXIT_ERROR,
    EXIT_RESTART,
    MODULE_MODE_FLAG,
    MODULE_MODE_SET,
)

w, h = os.get_terminal_size()


def banner_print(s: str) -> None:
    print("-" * w)
    print("[melobot 模块]：", s, sep="")
    print("-" * w)


if len(sys.argv) != 2:
    print("以模块模式运行时，必须有且只有一个命令行参数：bot 主程序路径")
    sys.exit(0)
bot_script_path = sys.argv[1]
cmd = [sys.executable, bot_script_path]
cwd = str(pathlib.Path.cwd().absolute().resolve(strict=True))
os.environ[MODULE_MODE_FLAG] = MODULE_MODE_SET

try:
    while True:
        ret = subprocess.run(
            cmd,
            env=os.environ,
            cwd=cwd,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        if ret.returncode == EXIT_CLOSE:
            break
        elif ret.returncode == EXIT_RESTART:
            banner_print("正在重启 bot 程序")
            continue
        elif ret.returncode == EXIT_ERROR:
            break
        else:
            banner_print("以模块运行模式运行时，bot 主程序返回了无法处理的返回值")
            break
except KeyboardInterrupt:
    pass



import asyncio as aio

from melobot.models.event import MessageEvent

from ..context.session import SESSION_LOCAL, BotSessionManager, any_event
from ..types.exceptions import *
from ..types.typing import *

if TYPE_CHECKING:
    from ..context.session import BotSession
    from ..models.event import MessageEvent
    from ..types.abc import (
        AbstractResponder,
        BotChecker,
        BotEvent,
        BotMatcher,
        BotParser,
        SessionRule,
    )
    from ..utils.logger import Logger


class EventHandler:
    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        plugin: Any,
        responder: "AbstractResponder",
        logger: "Logger",
        checker: "BotChecker",
        priority: PriorLevel,
        set_block: bool,
        temp: bool,
        session_rule: "SessionRule",
        session_hold: bool,
        direct_rouse: bool,
        conflict_wait: bool,
        conflict_cb: Callable[[], Coroutine],
    ) -> None:
        super().__init__()
        self.is_valid = True

        self.executor = executor
        self.responder = responder
        self.logger = logger
        self.checker = checker
        self.priority = priority
        self.set_block = set_block
        self.temp = temp

        self._run_lock = aio.Lock()
        self._rule = session_rule
        self._hold = session_hold
        self._plugin = plugin
        self._direct_rouse = direct_rouse
        self._conflict_cb = conflict_cb
        self._wait_flag = conflict_wait

        if session_rule is None:
            if session_hold or direct_rouse or conflict_wait or conflict_cb:
                raise EventHandlerError(
                    "使用 session_rule 参数后才能使用以下参数：session_hold， direct_rouse, \
                                      conflict_wait, conflict_callback"
                )

        if conflict_wait and conflict_cb:
            raise EventHandlerError(
                "参数 conflict_wait 为 True 时，冲突回调永远不会被调用"
            )

    async def _verify(self) -> bool:
        """
        验证事件是否有触发执行的资格（验权）
        """
        if self.checker:
            return await self.checker.check(any_event())
        return True

    async def _run_on_ctx(
        self,
        coro: Coroutine[Any, Any, T],
        session: "BotSession" = None,
        timeout: float = None,
    ) -> T:
        """
        在指定 session 上下文中运行协程。异常将会抛出
        """
        if session._handler is None:
            BotSessionManager.inject(session, self)
        try:
            s_token = SESSION_LOCAL._add_ctx(session)
            return await aio.wait_for(aio.create_task(coro), timeout=timeout)
        finally:
            SESSION_LOCAL._del_ctx(s_token)
            # 这里可能因 bot 停止运行，导致退出事件执行方法时 session 为挂起态。因此需要强制唤醒
            if session._hup_signal.is_set():
                BotSessionManager._rouse(session)

    async def _run(self, event: "BotEvent") -> None:
        """
        获取 session 然后准备运行 executor
        """
        try:
            session = None
            if not self._direct_rouse:
                res = await BotSessionManager.try_attach(event, self)
                if res:
                    return
            session = await BotSessionManager.get(event, self)
            # 如果因为冲突没有获得 session，且指定了冲突回调
            if session is None and self._conflict_cb:
                temp_session = BotSessionManager.make_temp(event)
                await self._run_on_ctx(self._conflict_cb(), temp_session)
            # 如果因为冲突没有获得 session，但没有冲突回调
            if session is None:
                return
            # 如果没有冲突，正常获得到了 session
            exec_coro = self.executor(self._plugin)
            await self._run_on_ctx(exec_coro, session)
        except DirectRetSignal:
            pass
        except Exception as e:
            executor_name = self.executor.__qualname__
            self.logger.error(
                f"插件 {self._plugin.ID} 事件处理方法 {executor_name} 发生异常"
            )
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_locals(locals()))
        finally:
            if session is None:
                return
            BotSessionManager.recycle(session, alive=self._hold)

    async def _pre_process(self, event: "BotEvent") -> Coroutine[Any, Any, bool]:
        session = BotSessionManager.make_temp(event)
        status = await self._run_on_ctx(self._verify(), session)
        return status

    async def evoke(self, event: "BotEvent") -> bool:
        """
        接收总线分发的事件的方法。返回是否决定处理的判断。
        便于 disptacher 进行优先级阻断。校验通过会自动处理事件。
        """
        if not self.is_valid:
            return False

        if not (await self._pre_process(event)):
            return False

        if not self.temp:
            aio.create_task(self._run(event))
            return True

        async with self._run_lock:
            if self.is_valid:
                aio.create_task(self._run(event))
                self.is_valid = False
                return True
            else:
                return False


class AllEventHandler(EventHandler):
    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        plugin: Any,
        responder: "AbstractResponder",
        logger: "Logger",
        checker: "BotChecker",
        priority: PriorLevel,
        set_block: bool,
        temp: bool,
        session_rule: "SessionRule",
        session_hold: bool,
        direct_rouse: bool,
        conflict_wait: bool,
        conflict_cb: Callable[[], Coroutine],
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb,
        )


class MsgEventHandler(EventHandler):
    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        plugin: Any,
        responder: "AbstractResponder",
        logger: "Logger",
        matcher: "BotMatcher" = None,
        parser: "BotParser" = None,
        checker: "BotChecker" = None,
        priority: PriorLevel = PriorLevel.MEAN,
        set_block: bool = False,
        temp: bool = False,
        session_rule: "SessionRule" = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb,
        )
        self.matcher = matcher
        self.parser = parser

        # matcher 和 parser 不能同时存在
        if self.matcher and self.parser:
            raise EventHandlerError("参数 matcher 和 parser 不能同时存在")

    def _match(
        self, event: "MessageEvent"
    ) -> bool | tuple[bool, str | None, ParseArgs | None]:
        """
        检查是否匹配
        """
        if self.matcher:
            return self.matcher.match(event.text)
        if self.parser:
            args_group = event._get_args(self.parser.id)
            if args_group == -1:
                args_group = self.parser.parse(event.text)
                event._store_args(self.parser.id, args_group)
            res, cmd_name, args = self.parser.test(args_group)
            return res, cmd_name, args
        return True

    async def _format(self, cmd_name: str, args: ParseArgs) -> bool:
        """
        格式化。只有 parser 存在时需要
        """
        if not self.parser.need_format:
            return True
        status = await self.parser.format(cmd_name, args)
        return status

    async def _pre_process(self, event: MessageEvent) -> Coroutine[Any, Any, bool]:
        session = BotSessionManager.make_temp(event)
        # 先进行 match，match 不支持回调，因此可以直接同步运行，而且无需异步加锁
        match_res = self._match(event)
        if isinstance(match_res, bool):
            if not match_res:
                return False
        else:
            res, cmd_name, args = match_res
            if not res:
                return False

        async def wrapped_func() -> bool:
            if not (await self._verify()):
                return False
            if self.parser:
                if not (await self._format(cmd_name, args)):
                    return False
            return True

        status = await self._run_on_ctx(wrapped_func(), session)
        return status


class ReqEventHandler(EventHandler):
    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        plugin: Any,
        responder: "AbstractResponder",
        logger: "Logger",
        checker: "BotChecker" = None,
        priority: PriorLevel = PriorLevel.MEAN,
        set_block: bool = False,
        temp: bool = False,
        session_rule: "SessionRule" = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb,
        )


class NoticeEventHandler(EventHandler):
    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        plugin: Any,
        responder: "AbstractResponder",
        logger: "Logger",
        checker: "BotChecker" = None,
        priority: PriorLevel = PriorLevel.MEAN,
        set_block: bool = False,
        temp: bool = False,
        session_rule: "SessionRule" = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb,
        )


class MetaEventHandler(EventHandler):
    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        plugin: Any,
        responder: "AbstractResponder",
        logger: "Logger",
        checker: "BotChecker" = None,
        priority: PriorLevel = PriorLevel.MEAN,
        set_block: bool = False,
        temp: bool = False,
        session_rule: "SessionRule" = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb,
        )


EVENT_HANDLER_MAP: dict[str, tuple[Type[EventHandler], ...]] = {
    "message": (MsgEventHandler, AllEventHandler),
    "request": (ReqEventHandler, AllEventHandler),
    "notice": (NoticeEventHandler, AllEventHandler),
    "meta": (MetaEventHandler, AllEventHandler),
}



import asyncio as aio
import inspect
import os
import sys

import better_exceptions

from .context.session import BotSessionManager
from .controller.dispatcher import BotDispatcher
from .controller.responder import BotResponder
from .io.forward_ws import ForwardWsConn
from .meta import EXIT_CLOSE, EXIT_RESTART, MODULE_MODE_FLAG, MODULE_MODE_SET, MetaInfo
from .models.event import BotEventBuilder
from .plugin.bot import BOT_PROXY, BotHookBus
from .plugin.handler import EVENT_HANDLER_MAP
from .plugin.ipc import PluginBus, PluginStore
from .plugin.plugin import Plugin, PluginLoader
from .types.exceptions import *
from .types.typing import *
from .utils.config import BotConfig
from .utils.logger import get_logger

if TYPE_CHECKING:
    from .utils.logger import Logger

better_exceptions.hook()

if sys.platform != "win32":
    import uvloop

    aio.set_event_loop_policy(uvloop.EventLoopPolicy())


class MeloBot:
    """MeloBot 类。负责 bot 的启动、加载和管理。"""

    def __init__(self) -> None:
        self.config: BotConfig
        self.info = MetaInfo()
        self._logger: "Logger" = None

        self.life: aio.Task = None
        self.plugins: Dict[str, Plugin] = {}
        self.loader = PluginLoader

        self.linker: ForwardWsConn
        self.responder: BotResponder
        self.dispatcher: BotDispatcher
        self.ctx_manager = BotSessionManager
        self.event_builder = BotEventBuilder
        self.plugin_bus = PluginBus
        self.plugin_store = PluginStore
        self.bot_bus = BotHookBus
        self.proxy = BOT_PROXY

        self.__init_flag__ = False
        self.exit_code = EXIT_CLOSE

    @property
    def loop(self) -> aio.AbstractEventLoop:
        return aio.get_running_loop()

    @property
    def slack(self) -> bool:
        return self.linker.slack

    @slack.setter
    def slack(self, value: bool) -> None:
        self.linker.slack = value

    @property
    def logger(self) -> "Logger":
        return self._logger if self._logger else get_logger()

    def init(self, config_dir) -> None:
        """
        为 bot 核心加载配置文件
        """
        if self.__init_flag__:
            self.logger.error("bot 不能重复初始化")
            sys.exit(self.exit_code)

        self.config = BotConfig(config_dir)
        self._logger = get_logger(self.config.log_dir_path, self.config.log_level)
        self.linker = ForwardWsConn(
            self.config.connect_host,
            self.config.connect_port,
            self.config.max_conn_try,
            self.config.conn_try_interval,
            self.config.cooldown_time,
            self.event_builder,
            self.bot_bus,
            self.logger,
        )
        self.responder = BotResponder(self.logger)
        self.dispatcher = BotDispatcher(
            EVENT_HANDLER_MAP, self.bot_bus, self.ctx_manager, self.logger
        )
        self.ctx_manager._bind(self.responder)
        self.plugin_bus._bind(self.logger)
        self.bot_bus._bind(self.logger)

        self.logger.info("欢迎使用 melobot v2")
        self.logger.info(f"melobot 在 AGPL3 协议下开源发行")
        self.logger.info(f"运行版本：{self.info.VER}，平台：{self.info.PLATFORM}")
        if os.environ.get(MODULE_MODE_FLAG) == MODULE_MODE_SET:
            self.logger.info("当前运行模式：模块运行模式")
        else:
            self.logger.info("当前运行模式：脚本运行模式")
        self.logger.info("bot 核心初始化完成")
        self.__init_flag__ = True

    def load_plugin(self, plugin_target: str | Type[Plugin]) -> None:
        """
        为 bot 加载运行插件。支持传入插件起始目录字符串（绝对路径）或插件类
        """
        if not self.__init_flag__:
            self.logger.error("加载插件必须在初始化之后进行")
            sys.exit(self.exit_code)

        plugin_dir = (
            inspect.getfile(plugin_target)
            if not isinstance(plugin_target, str)
            else plugin_target
        )
        self.logger.debug(f"尝试加载来自 {plugin_dir} 的插件")
        plugin = self.loader.load(plugin_target, self.logger, self.responder)
        exist_plugin = self.plugins.get(plugin.ID)
        if exist_plugin is None:
            self.plugins[plugin.ID] = plugin
            self.dispatcher.add_handlers(plugin._handlers)
            self.logger.info(f"成功加载插件：{plugin.ID}")
        else:
            self.logger.error(
                f"加载插件出错：插件名称重复, 尝试加载：{plugin_dir}，已加载：{exist_plugin.plugin_dir}"
            )
            sys.exit(self.exit_code)

    def load_plugins(self, plugins_dir: str) -> None:
        """
        从插件目录加载多个 bot 运行插件。但必须遵循标准插件格式
        """
        items = os.listdir(plugins_dir)
        for item in items:
            path = os.path.join(plugins_dir, item)
            if os.path.isdir(path) and os.path.basename(path) != "__pycache__":
                self.load_plugin(path)

    async def _run(self) -> None:
        if len(self.plugins) == 0:
            self.logger.warning("没有加载任何插件，bot 将不会有任何操作")

        await self.bot_bus.emit(BotLife.LOADED)
        self.responder.bind(self.linker)
        self.linker.bind(self.dispatcher, self.responder)
        try:
            async with self.linker:
                aio.create_task(self.linker.send_queue_watch())
                self.life = aio.create_task(self.linker.listen())
                self.proxy._bind(self)
                self.logger.info("bot 开始正常运行")
                await self.life
        except Exception as e:
            self.logger.error(f"bot 核心无法继续运行。异常：{e}")
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_locals(locals()))
        finally:
            await self.bot_bus.emit(BotLife.BEFORE_STOP, wait=True)
            self.logger.info("bot 已清理运行时资源")
            sys.exit(self.exit_code)

    def run(self) -> None:
        """
        运行 bot
        """
        if not self.__init_flag__:
            self.logger.error("必须先初始化才能启动")
            sys.exit(self.exit_code)
        try:
            """
            一定要手动设置事件循环。在启动前，部分核心模块初始化异步对象，
            已经生成了事件循环，不设置而直接使用 asyncio.run 将会运行一个新的事件循环，
            不同事件循环的异步对象无法直接通信
            """
            aio.set_event_loop(aio.get_event_loop())
            # 使用 asyncio.run 可以保证发生各种异常时一定取消所有任务（包括键盘中断）
            aio.run(self._run())
        except KeyboardInterrupt:
            pass

    async def close(self) -> None:
        """
        停止 bot
        """
        if self.life is None:
            self.logger.error("bot 尚未运行，无需停止")
            sys.exit(self.exit_code)
        await self.bot_bus.emit(BotLife.BEFORE_CLOSE, wait=True)
        self.life.cancel()

    async def restart(self) -> None:
        """
        重启 bot
        """
        if self.life is None:
            self.logger.error("bot 尚未运行，无需重启")
            sys.exit(self.exit_code)
        if os.environ.get(MODULE_MODE_FLAG) != MODULE_MODE_SET:
            self.logger.error("只有在模块运行模式下，才能使用 bot 重启功能")
            return
        self.logger.info("bot 即将进行重新启动...")
        self.exit_code = EXIT_RESTART
        self.life.cancel()
